/*
 * Main File including the main-routine with initializations-
 * and stack-routines.
 *
 * Author: Simon Kueppers
 * Email: simon.kueppers@web.de
 * Homepage: http://klinkerstein.m-faq.de
 * */

#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "Hardware/Enc28j60.h"
#include "Hardware/Spi.h"
#include "Hardware/Uart.h"
#include "Net/uip/uip.h"
#include "Net/uip/uip_arp.h"
#include "Net/uip/uip_TcpAppHub.h"
#include "Net/uip/uip_UdpAppHub.h"
#include <avr/eeprom.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "ports.h"
#include "options.h"
#include "screens.h"

#define UART_BAUD_RATE      9600

#define CUR_VER		0x01

#define EE_IP		0
#define EE_NM		4
#define EE_GW		8
#define EE_DC		12
#define EE_VER		15
// reserved 16-63
#define EE_IO		64
#define EE_LCD_C	89
#define EE_LCD_D	90
#define EE_OPT_C	91
// reserver 92, 93, 94, 95
#define EE_M		96
// reserved 97-127
#define EE_OPTS		128
// 48 Roule entries (8)
#define EE_LBLS		512

#define EE_END 		2047

extern tGpio g_Gpio[];
extern tLCDPins g_LCDPins[];
extern tVPort g_VPort[];
extern uint8_t g_LCDSupport;
extern uint8_t g_LCDLines;
extern uint8_t g_LCDAuto;
extern tOption* g_Option;
extern tScreen* g_Screen;

char buffer[128];
int bufferlen = 0;
uip_ipaddr_t IpAddr;

static uint8_t g_nPrescaler = 100;
static volatile struct
{
	//uint8_t fPeriodic :1;
	uint8_t fPeriodic;
} g_nFlags =
{ 0 };

void Reboot()
{
	cli(); // Disable all interrupts

	do {
	  for(;;){}
	} while(0);
}

void uartprint(const char * _Format, ...)
{
	va_list argptr;
	va_start(argptr, _Format);
	vsnprintf(buffer, 128, _Format, argptr);
	va_end(argptr);

	uart_puts(buffer);
}

void uartprint_P(const prog_char * _Format, ...)
{
	va_list argptr;
	va_start(argptr, _Format);
	vsnprintf_P(buffer, 128, _Format, argptr);
	va_end(argptr);

	uart_puts(buffer);
}

#define SPLASH		PSTR("Starting uWebSrv on port %d\r\n")
#define IP 			PSTR("IP  %d.%d.%d.%d\r\n")
#define MASK 		PSTR("Netmask %d.%d.%d.%d\r\n")
#define GW 			PSTR("Gateway %d.%d.%d.%d\r\n")
#define DHCP_ON 	PSTR("DHCP On\r\n")
#define DHCP_OFF	PSTR("DHCP Off\r\n")
#define DHCP_RE		PSTR("DHCP Renew\r\n")
#define DHCP_LEASE	PSTR("DHCP Lease:\r\n")
#define DHCP_EX		PSTR("Expires in %ld seconds\r\n")
#define USAGE		PSTR("Unknown Command\r\n"\
						"\r\n"\
						"\tGet Setup: Setup\r\n"\
						"\tGet Ip Config: Config\r\n"\
						"\tGet DHCP: DHCP On/Off/Renew\r\n"\
						"\tSet Host IP Address: SetIP a.b.c.d\r\n"\
						"\tSet Network Mask: SetNM a.b.c.d\r\n"\
						"\tSet Gateway: SetGW a.b.c.d\r\n"\
						"\r\n")


int ReadIP(char* buffer, uip_ipaddr_t *__a)
{
    int Ip[4];
	if(sscanf_P(buffer,PSTR("%d.%d.%d.%d"),&Ip[0],&Ip[1],&Ip[2],&Ip[3]) == 4)
	{
		uip_ipaddr(*__a, Ip[0], Ip[1], Ip[2], Ip[3]);
		return 1;
	}

	uartprint_P(PSTR("Bad IP\r\n"));
	return 0;
}

void eeprom_read_IP4 (const uint16_t __p, uip_ipaddr_t *__a)
{
	uip_ipaddr(*__a,
			eeprom_read_byte((uint8_t*)__p),
			eeprom_read_byte((uint8_t*)(__p+1)),
			eeprom_read_byte((uint8_t*)(__p+2)),
			eeprom_read_byte((uint8_t*)(__p+3)));
}

void eeprom_write_IP4 (const uint16_t __p, uip_ipaddr_t *__a)
{
	eeprom_write_byte((uint8_t*)__p,uip_ipaddr1(*__a));
	eeprom_write_byte((uint8_t*)(__p+1),uip_ipaddr2(*__a));
	eeprom_write_byte((uint8_t*)(__p+2),uip_ipaddr3(*__a));
	eeprom_write_byte((uint8_t*)(__p+3),uip_ipaddr4(*__a));
}

void PrintSetup()
{
	if(eeprom_read_byte((uint8_t*)EE_DC))
		uart_puts_p(DHCP_ON);
	else
		uart_puts_p(DHCP_OFF);

	eeprom_read_IP4(EE_IP, &IpAddr);
	uartprint_P(IP, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

	eeprom_read_IP4(EE_NM, &IpAddr);
	uartprint_P(MASK, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

	eeprom_read_IP4(EE_GW, &IpAddr);
	uartprint_P(GW, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
}

void PrintConfig()
{
	uip_gethostaddr(IpAddr);
	uartprint_P(IP, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

	uip_getnetmask(IpAddr);
	uartprint_P(MASK, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

	uip_getdraddr(IpAddr);
	uartprint_P(GW, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
}

void dhcpc_configured(const struct dhcpc_state *s)
{
  uip_sethostaddr(s->ipaddr);
  uip_setnetmask(s->netmask);
  uip_setdraddr(s->default_router);
  //resolv_conf(s->dnsaddr);

  uart_puts_p(DHCP_LEASE);

  PrintConfig();

  uartprint_P(DHCP_EX, ntohs(s->lease_time[0])*65536ul + ntohs(s->lease_time[1]));
}

char g_EE_Buff[21];
void Load()
{
	if(eeprom_read_byte((uint8_t*)EE_VER) != CUR_VER)
		return;

	int i,j,k,l;
	k=0;
	for(i=0; i<25;i++)
	{
		g_Gpio[i].IO = eeprom_read_byte((uint8_t*)EE_IO+i);
		if(g_Gpio[i].IO >= LC_RS && g_Gpio[i].IO <= LC_SR)
			g_LCDPins[g_Gpio[i].IO-LC_RS].Port = i;

		for (j=0; (g_EE_Buff[j] = eeprom_read_byte((uint8_t*)EE_LBLS+k++)) != 0 && EE_LBLS+k < EE_END;j++);

		SetPortLabel(i,g_EE_Buff);
	}

	uint8_t LCD = eeprom_read_byte((uint8_t*)EE_LCD_C);
	g_LCDSupport = LCD & 0x0F;
	g_LCDLines = (LCD & 0xF0) >> 4;
	g_LCDAuto = eeprom_read_byte((uint8_t*)EE_LCD_D);

	for(i=0; i<16;i++)
	{
		g_VPort[i].IO = eeprom_read_byte((uint8_t*)EE_M+i);

		for (j=0; (g_EE_Buff[j] = eeprom_read_byte((uint8_t*)EE_LBLS+k++)) != 0 && EE_LBLS+k < EE_END;j++);

		SetPortLabel(25+i,g_EE_Buff);
	}

	int OptionCnt = eeprom_read_byte((uint8_t*)EE_OPT_C);
	for(i=0; i<OptionCnt;i++)
	{
		for(j=0; j<8;j++)
		{
			g_EE_Buff[j] = eeprom_read_byte((uint8_t*)(EE_OPTS+(i*8)+j));
		}
		AddOption((uint8_t*)g_EE_Buff);
	}

	for (i=j=k=l=0; (g_EE_Buff[l] = eeprom_read_byte((uint8_t*)EE_END-k++)) != 0xFF && EE_END-k > EE_LBLS;)
	{
		if(g_EE_Buff[l] == 0)
		{
			SetScr(i,j,g_EE_Buff);
			j++;
			if(j>=4)
			{
				SetScrAction(i, eeprom_read_byte((uint8_t*)EE_END-k++));
				i++;
				j=0;
			}
			l=0;
		}
		else
			l++;
	}
}

void Save(int n)
{
	if(eeprom_read_byte((uint8_t*)EE_VER) != CUR_VER)
	{
		n = 4;
		eeprom_write_byte((uint8_t*)EE_VER, CUR_VER);
	}

	if(n == 0 || n == 4)
	{
		int i,j,k;
		k=0;
		for(i=0; i<25;i++)
		{
			eeprom_write_byte((uint8_t*)EE_IO+i, g_Gpio[i].IO);

			if(g_Gpio[i].Label)
			{
				for(j=0;g_Gpio[i].Label[j];j++)
				{
					eeprom_write_byte((uint8_t*)(EE_LBLS+k),g_Gpio[i].Label[j]);
					k++;
				}
			}
			eeprom_write_byte((uint8_t*)(EE_LBLS+k),0);
			k++;
		}

		uint8_t LCD = (g_LCDSupport & 0xF) | ((g_LCDLines << 4) & 0xF0);
		eeprom_write_byte((uint8_t*)EE_LCD_C, LCD);
		eeprom_write_byte((uint8_t*)EE_LCD_D, g_LCDAuto);

		for(i=0; i<16;i++)
		{
			eeprom_write_byte((uint8_t*)EE_M+i, g_VPort[i].IO);

			if(g_VPort[i].Label)
			{
				for(j=0;g_VPort[i].Label[j];j++)
				{
					eeprom_write_byte((uint8_t*)(EE_LBLS+k),g_VPort[i].Label[j]);
					k++;
				}
			}
			eeprom_write_byte((uint8_t*)(EE_LBLS+k),0);
			k++;
		}
	}

	if(n == 1 || n == 4)
	{
		int i,j;
		int OptionCnt = GetOptCount();
		if(OptionCnt > 48)
			OptionCnt = 48;
		eeprom_write_byte((uint8_t*)EE_OPT_C, OptionCnt);
		for(i=0; i<OptionCnt;i++)
		{
			memcpy(g_EE_Buff,&g_Option[i],8);
			for(j=0; j<8;j++)
			{
				eeprom_write_byte((uint8_t*)(EE_OPTS+(i*8)+j),g_EE_Buff[j]);
			}
		}
	}

	if(n == 2 || n == 4)
	{
		int i,j,k,l;
		k=0;
		for(i=0; i<GetScrCount();i++)
		{
			for(j=0;j<4;j++)
			{
				char* Line = GetScr(i,j);
				for(l=0;Line[l] != 0;l++)
				{
					eeprom_write_byte((uint8_t*)(EE_END-k),Line[l]);
					k++;
				}
				eeprom_write_byte((uint8_t*)(EE_END-k),0);
				k++;
			}
			eeprom_write_byte((uint8_t*)(EE_END-k),GetScrAction(i));
			k++;
		}

		eeprom_write_byte((uint8_t*)(EE_END-k),0xFF);
	}
}

int main()
{
	unsigned int ret;
	uint8_t nSL = 0;
	uint8_t nSR = 0;

	SpiInit();
	uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );
	Enc28j60Init();
	uip_arp_init();
	uip_init();
	uip_UdpAppHubInit();
	uip_TcpAppHubInit();
	Enc28j60SetClockPrescaler(2);

	//TCCR1B = (1<<WGM12)|(3<<CS10);
	//OCR1A = 25000; //16000000 / 64 / 25000 = 10Hz (100ms)
    //TIMSK1 |= (1 << OCIE1A);

	TCCR0A = (1<<WGM01);
	TCCR0B = (5<<CS00);
	OCR0A = 156; //16000000 / 1024 / 156 = 100Hz (10ms)
	TIMSK0 |= (1 << OCIE0A);

    if(eeprom_read_byte((uint8_t*)EE_DC))
    {
    	dhcpc_request();
    }
    else
    {
    	eeprom_read_IP4(EE_IP, &IpAddr);
		uip_sethostaddr(IpAddr);

		eeprom_read_IP4(EE_NM, &IpAddr);
		uip_setnetmask(IpAddr);

		eeprom_read_IP4(EE_GW, &IpAddr);
		uip_setdraddr(IpAddr);;
    }

	sei	();

	Load();

	InitScreen();

    uartprint_P(SPLASH,PORT_HTTPD);

    wdt_enable(WDTO_2S);

	while (1)
	{
		wdt_reset();

		uip_len = Enc28j60Receive(uip_buf);

		if (uip_len> 0)
		{
			if (((struct uip_eth_hdr *)&uip_buf[0])->type
					== htons(UIP_ETHTYPE_IP))
			{
				//uip_arp_ipin();
				uip_input();

				if (uip_len> 0)
				{
					uip_arp_out();
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}

			}
			else if (((struct uip_eth_hdr *)&uip_buf[0])->type
					== htons(UIP_ETHTYPE_ARP))
			{
				uip_arp_arpin();
				if (uip_len> 0)
				{
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}
			}
		}

		if (g_nFlags.fPeriodic >= 10)
		//if (g_nFlags.fPeriodic)
		{
			cli();
			g_nFlags.fPeriodic = 0;
			sei();

			int i = UIP_CONNS;
			while (i)
			{
				i--;
				uip_periodic(i);
				if (uip_len> 0)
				{
					uip_arp_out();
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}
			}

#if UIP_UDP
			int j = UIP_UDP_CONNS;
			while (j)
			{
				j--;
				uip_udp_periodic(j);
				if (uip_len> 0)
				{
					uip_arp_out();
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}
			}
#endif /* UIP_UDP */

			g_nPrescaler--;

			if(!nSL && GetLCDPort(SL) != NONE && GetBit(GetLCDPort(SL)))
			{
				nSL = 1;
				if(!(GetLCDPort(SR) != NONE && GetBit(GetLCDPort(SR))))
					nSR = 0;
			}
			if(!nSR &&GetLCDPort(SR) != NONE && GetBit(GetLCDPort(SR)))
			{
				nSR = 1;
				if(!(GetLCDPort(SL) != NONE && GetBit(GetLCDPort(SL))))
					nSL = 0;
			}

			if (g_nPrescaler%5 == 0)
			{
				//Every 0.5 seconds
				SwitchPage(nSL,nSR);
				nSL = nSR = 0;
			}

			if (g_nPrescaler%10 == 0)
			{
				//Every 1 seconds
				MgmtDo();
				dhcpc_timer();
			}

			if (g_nPrescaler == 0)
			{
				//Every 10 seconds
				uip_arp_timer();
				g_nPrescaler = 100;
			}
		}

		ret = uart_getc();
		if ( (ret & UART_NO_DATA) == 0 )
		{
			if ( ret & UART_FRAME_ERROR )
			{
				uart_puts_P("UART Frame Error\r\n");
			}
			if ( ret & UART_OVERRUN_ERROR )
			{
				uart_puts_P("UART Overrun Error\r\n");
			}
			if ( ret & UART_BUFFER_OVERFLOW )
			{
				uart_puts_P("Buffer overflow error\r\n");
			}

			if((char)ret == 8 && bufferlen > 0)
			{
				uart_putc( (unsigned char)ret );
				bufferlen--;
				buffer[bufferlen] = '\0';
			}
			else if((char)ret != '\r' && (char)ret != '\n')
			{
				uart_putc( (unsigned char)ret );
				buffer[bufferlen] = (char)ret;
				bufferlen++;
			}
			else if((char)ret == '\r' && bufferlen > 0)
			{
				uart_puts_P("\r\n");
				buffer[bufferlen] = '\0';
				bufferlen = 0;

				if(strncmp_P(buffer,PSTR("Setup"),5) == 0)
				{
					PrintSetup();
				}
				else if(strncmp_P(buffer,PSTR("Config"),6) == 0)
				{
					PrintConfig();
				}
				else if(strncmp_P(buffer,PSTR("DHCP"),4) == 0)
				{
					if(strncmp_P(buffer+5,PSTR("On"),2) == 0)
					{
						eeprom_write_byte((uint8_t*)EE_DC, 1);
						uart_puts_p(DHCP_ON);
					}
					else if(strncmp_P(buffer+5,PSTR("Off"),3) == 0)
					{
						eeprom_write_byte((uint8_t*)EE_DC, 0);
						uart_puts_p(DHCP_OFF);
					}
					else if(strncmp_P(buffer+5,PSTR("Renew"),5) == 0)
					{
						uart_puts_p(DHCP_RE);
						dhcpc_request();
					}
				}
				else if(strncmp_P(buffer,PSTR("SetIP"),5) == 0)
				{
					if(ReadIP(buffer+6,&IpAddr))
					{
						uip_sethostaddr(IpAddr);
						eeprom_write_IP4(EE_IP, &IpAddr);
						uartprint_P(IP,uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
					}
				}
				else if(strncmp_P(buffer,PSTR("SetNM"),5) == 0)
				{
					if(ReadIP(buffer+6,&IpAddr))
					{
						uip_setnetmask(IpAddr);
						eeprom_write_IP4(EE_NM, &IpAddr);
						uartprint_P(MASK,uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
					}
				}
				else if(strncmp_P(buffer,PSTR("SetGW"),5) == 0)
				{
					if(ReadIP(buffer+6,&IpAddr))
					{
						uip_setdraddr(IpAddr);
						eeprom_write_IP4(EE_GW, &IpAddr);
						uartprint_P(GW,uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
					}
				}
				else if(strncmp_P(buffer,PSTR("Reboot"),6) == 0)
				{
					Reboot();
				}
				else
				{
					uart_puts_p(USAGE);
				}
			}
		}
	}

	return 0;
}

//ISR(TIMER1_COMPA_vect)
ISR(TIMER0_COMPA_vect)
{
	//g_nFlags.fPeriodic = 1;
	g_nFlags.fPeriodic++;
}
